"""Internal functionality for validating role definitions."""

__all__ = ["validate_roles"]

from viv_compiler import config, errors, external_types, utils


def validate_roles(construct_definition: external_types.ConstructDefinition) -> None:
    """Validate the role definitions in the given construct definition.

    Args:
        construct_definition: A construct definition in a compiled content bundle.

    Returns:
        None. Only returns if no issue was detected downstream.
    """
    # Validate the role labels
    _validate_role_labels(construct_definition=construct_definition)
    # Make sure all roles have proper 'min' and 'max' values
    _validate_role_min_and_max_values(construct_definition=construct_definition)
    # Make sure all roles have proper 'chance' and 'mean' values
    _validate_role_chance_and_mean_values(construct_definition=construct_definition)
    # Make sure all roles with casting pools have proper ones
    _validate_casting_pool_directives(construct_definition=construct_definition)
    # If this is a construct of a type using an initiator role, validate its initiator role
    if construct_definition['type'] in config.CONSTRUCT_TYPES_USING_INITIATOR_ROLES:
        _validate_initiator_role(construct_definition=construct_definition)


def _validate_role_labels(construct_definition: external_types.ConstructDefinition) -> None:
    """Ensure that the given construct definition has valid role labels.

    Args:
        construct_definition: A construct definition in a compiled content bundle.

    Returns:
        None. Only returns if no issue is encountered.

    Raises:
        VivCompileError: The construct definition does not have valid role labels.
    """
    for role_definition in construct_definition['roles'].values():
        # Ensure that the labels are permitted given the construct type
        reconstructed_role_labels = _reconstruct_role_labels(role_definition=role_definition)
        valid_role_labels = config.ROLE_LABELS_PERMITTED_IN_CONSTRUCT_OF_TYPE[construct_definition["type"]]
        for role_label in reconstructed_role_labels:
            if role_label not in valid_role_labels:
                raise errors.VivCompileError(
                    f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} '{construct_definition['name']}' "
                    f"has role '{role_definition['name']}' with invalid role label '{role_label}' (only these labels "
                    f"are allowed in a {utils.CONSTRUCT_LABEL[construct_definition['type']].lower()}: "
                    f"{', '.join(valid_role_labels)})"
                )
        # Ensure that only character roles have a role-participation mode attributed
        if role_definition['participationMode']:
            if role_definition['entityType'] != external_types.RoleEntityType.CHARACTER:
                raise errors.VivCompileError(
                    f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} '{construct_definition['name']}' "
                    f" has role '{role_definition['name']}' with '{role_definition['participationMode']}' "
                    f"label but no 'character' label ('{role_definition['participationMode']}' may only be "
                    f"used for character roles)"
                )
        # Ensure that the `precast` label is used properly
        if role_definition['precast'] and not utils.is_initiator_role(role_definition=role_definition):
            if (
                construct_definition['type'] == external_types.ConstructDiscriminator.ACTION or
                construct_definition['type'] == external_types.ConstructDiscriminator.ACTION_SELECTOR
            ):
                if not construct_definition['reserved']:
                    raise errors.VivCompileError(
                        f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} '{construct_definition['name']}' "
                        f" is not marked 'reserved' and has role {role_definition['name']}' with 'precast' label "
                        f"(only allowed in reserved actions/selectors)"
                    )
        # Ensure that no incompatible labels appear
        for pool in config.MUTUALLY_INCOMPATIBLE_ROLE_LABELS:
            for field in pool:
                if field not in reconstructed_role_labels:
                    continue
                for other_field in pool:
                    if field == other_field:
                        continue
                    if other_field not in reconstructed_role_labels:
                        continue
                    raise errors.VivCompileError(
                        f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} "
                        f"'{construct_definition['name']}' has role '{role_definition['name']}' with incompatible "
                        f"(implied) role labels: '{field}' and '{other_field}'"
                    )


def _reconstruct_role_labels(role_definition: external_types.RoleDefinition) -> list[str]:
    """Return a list containing role labels reconstructed for the given role.

    Args:
        role_definition: Definition for the role whose labels will be reconstructed.

    Returns:
        A list containing role labels reconstructed for the given role.
    """
    reconstructed_role_labels: list[str] = []
    match role_definition['entityType']:
        case external_types.RoleEntityType.CHARACTER:
            reconstructed_role_labels.append("character")
        case external_types.RoleEntityType.ITEM:
            reconstructed_role_labels.append("item")
        case external_types.RoleEntityType.LOCATION:
            reconstructed_role_labels.append("location")
        case external_types.RoleEntityType.ACTION:
            reconstructed_role_labels.append("action")
        case external_types.RoleEntityType.SYMBOL:
            reconstructed_role_labels.append("symbol")
        case _:
            raise errors.VivCompileError(
                f"Unexpected entity type for role '{role_definition['name']}': '{role_definition['entityType']}'"
            )
    match role_definition['participationMode']:
        case None:
            pass
        case external_types.RoleParticipationMode.INITIATOR:
            reconstructed_role_labels.append("initiator")
        case external_types.RoleParticipationMode.PARTNER:
            reconstructed_role_labels.append("partner")
        case external_types.RoleParticipationMode.RECIPIENT:
            reconstructed_role_labels.append("recipient")
        case external_types.RoleParticipationMode.BYSTANDER:
            reconstructed_role_labels.append("bystander")
    if role_definition["anywhere"]:
        reconstructed_role_labels.append("anywhere")
    if role_definition["precast"]:
        reconstructed_role_labels.append("precast")
    if role_definition["spawn"]:
        reconstructed_role_labels.append("spawn")
    return reconstructed_role_labels


def _validate_role_min_and_max_values(construct_definition: external_types.ConstructDefinition) -> None:
    """Ensure that the given construct definition has no role definition with invalid `min` or `max` value.

    Args:
        construct_definition: A construct definition in a compiled content bundle.

    Returns:
        None. Only returns if no issue is encountered.

    Raises:
        VivCompileError: The construct definition has a role with invalid `min` or `max` value.
    """
    for role_definition in construct_definition["roles"].values():
        # Detect minimums less than 0
        if role_definition['min'] < 0:
            raise errors.VivCompileError(
                f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} '{construct_definition['name']}' "
                f"has role '{role_definition['name']}' with negative min (must be 0 or greater)"
            )
        # Detect maximums of 0 or less
        if role_definition['max'] < 1:
            raise errors.VivCompileError(
                f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} '{construct_definition['name']}' "
                f"has role '{role_definition['name']}' with max less than 1 (to turn off role, "
                f"comment it out or use chance of [0%])"
            )
        # Detect role-count minimums that are greater than role-count maximums
        if role_definition['min'] > role_definition['max']:
            raise errors.VivCompileError(
                f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} '{construct_definition['name']}' "
                f"has role '{role_definition['name']}' with min > max"
            )


def _validate_role_chance_and_mean_values(construct_definition: external_types.ConstructDefinition) -> None:
    """Ensure that the given construct definition has no role definition with invalid `chance` or `mean` value.

    Args:
        construct_definition: A construct definition in a compiled content bundle.

    Returns:
        None. Only returns if no issue is encountered.

    Raises:
        VivCompileError: The construct definition has a role with invalid `chance` or `mean` value.
    """
    for role_definition in construct_definition["roles"].values():
        if role_definition['chance']:
            # Confirm that 'chance' is only present if there are optional slots for this role
            if role_definition['min'] == role_definition['max']:
                raise errors.VivCompileError(
                    f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} '{construct_definition['name']}' "
                    f"has role '{role_definition['name']}' with optional-slot casting probability, but this is "
                    f"only allowed for roles with optional slots (max > min)"
                )
            # Confirm that 'chance', if present, is between 0.0 and 1.0
            if role_definition['chance'] < 0.0 or role_definition['chance'] > 1.0:
                raise errors.VivCompileError(
                    f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} '{construct_definition['name']}' "
                    f"has role '{role_definition['name']}' with invalid optional-slot casting probability "
                    f" (must be between 0-100%)"
                )
        # Confirm that 'mean', if present, is between min and max
        if role_definition['mean']:
            role_min, role_max = role_definition['min'], role_definition['max']
            if role_definition['mean'] < role_min or role_definition['mean'] > role_max:
                raise errors.VivCompileError(
                    f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} '{construct_definition['name']}' "
                    f"has role '{role_definition['name']}' with invalid slots mean (must be between min and max)"
                )


def _validate_casting_pool_directives(construct_definition: external_types.ConstructDefinition) -> None:
    """Ensure that the given construct definition does not have an invalid casting-pool directive.

    Note: During preliminary validation we would have already flagged the following issues:
        - A casting-pool directive that references an undefined role or an optional role.
        - An initiator role with a casting-pool directive.

    Args:
        construct_definition: A construct definition in a compiled content bundle.

    Returns:
        None. Only returns if no issue is encountered.

    Raises:
        VivCompileError: The construct definition has a role with a missing casting-pool declaration.
        VivCompileError: The construct definition has a role with an invalid casting-pool declaration.
    """
    for role_definition in construct_definition["roles"].values():
        requires_pool_directive = False
        if role_definition['precast'] or role_definition['spawn']:
            pass
        elif role_definition['entityType'] == external_types.RoleEntityType.ACTION:
            requires_pool_directive = True
        elif role_definition['entityType'] == external_types.RoleEntityType.SYMBOL:
            requires_pool_directive = True
        if requires_pool_directive and not role_definition["pool"]:
            raise errors.VivCompileError(
                f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} '{construct_definition['name']}' "
                f"has role '{role_definition['name']}' that requires a pool declaration but does not have one"
            )
        # Force "is" casting-pool directives to correspond to a max of exactly 1, since this is the "is"
        # semantics. Note that we don't actually retain any data marking that the 'is' operator was used, but
        # rather we create a casting-pool expression that is a list containing a single element, that being the
        # reference for the entity associated with the 'is' usage. We could end up with this same shape if the
        # author uses 'from' with a literal singleton array -- e.g. "hat from ["hat"]: symbol" -- so the error
        # message will refer to both potential causes.
        if role_definition['pool']:
            if role_definition["pool"]["body"]["type"] == external_types.ExpressionDiscriminator.LIST:
                if len(role_definition["pool"]["body"]["value"]) == 1:
                    # If we're in here, this is either an 'is' usage or a case of 'from' with a literal singleton array
                    if role_definition['max'] != 1:
                        raise errors.VivCompileError(
                            f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} "
                            f"'{construct_definition['name']}' has role '{role_definition['name']}' that uses 'is' "
                            f"casting-pool directive (or 'from' with a singleton array) with a max other than 1"
                        )
            # Prohibit casting-pool directives for roles carrying the 'precast' label
            if role_definition['precast']:
                raise errors.VivCompileError(
                    f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} '{construct_definition['name']}' "
                    f" has role '{role_definition['name']}' that carries the 'precast' label but has a casting-pool "
                    f"directive (these are incompatible, because the directive would always be ignored)"
                )
            # Prohibit casting-pool directives for roles carrying the `spawn` label
            if role_definition['spawn']:
                raise errors.VivCompileError(
                    f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} '{construct_definition['name']}' "
                    f" has role '{role_definition['name']}' that carries the 'spawn' label but has a casting-pool "
                    f"directive (these are incompatible)"
                )
            # Prohibit casting-pool directives anchored in roles carrying the `spawn` label
            if role_definition['pool']:
                for anchor_role_name in utils.get_all_referenced_roles(ast_chunk=role_definition["pool"]):
                    anchor_role_definition = construct_definition['roles'][anchor_role_name]
                    if anchor_role_definition['spawn']:
                        raise errors.VivCompileError(
                            f"Casting-pool directive for role '{role_definition['name']}' in "
                            f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} "
                            f"'{construct_definition['name']}' references a role carrying the 'spawn' label"
                        )


def _validate_initiator_role(construct_definition: external_types.ActionDefinition) -> None:
    """Ensure that the given construct definition has a valid initiator role definition.

    Note: We already prevalidated that the construct definition has exactly one initiator role.

    Args:
        construct_definition: A construct definition from a compiled content bundle.

    Returns:
        None. Only returns if no issue is encountered.

    Raises:
        VivCompileError: The construct definition has an invalid initiator role definition.
    """
    # Retrieve the definition for the initiator role
    initiator_role_definition = construct_definition['roles'][construct_definition['initiator']]
    # Make sure that the initiator role casts exactly one entity
    if initiator_role_definition['min'] != 1:
        raise errors.VivCompileError(
            f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} '{construct_definition['name']}' "
            f"has initiator role '{initiator_role_definition['name']}' with min other than 1 "
            f" (there must be a single initiator)"
        )
    if initiator_role_definition['max'] != 1:
        raise errors.VivCompileError(
            f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} '{construct_definition['name']}' "
            f"has initiator role '{initiator_role_definition['name']}' with max other than 1 "
            f"(there must be a single initiator)"
        )
    # Make sure that the initiator role has no specified casting mean
    if initiator_role_definition['mean']:
        raise errors.VivCompileError(
            f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} '{construct_definition['name']}' "
            f"has initiator role '{initiator_role_definition['name']}' with declared casting mean "
            f"(there must be a single initiator)"
        )
    # Make sure that the initiator role has no specified casting chance
    if initiator_role_definition['chance']:
        raise errors.VivCompileError(
            f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} '{construct_definition['name']}' "
            f"has initiator role '{initiator_role_definition['name']}' with declared casting chance "
            f"(there must be a single initiator)"
        )



"""Mixin class containing visitor methods associated with Viv temporal constraints."""

__all__ = ["VisitorMixinTemporalConstraints"]

from typing import cast

from arpeggio import NonTerminal, PTNodeVisitor

from viv_compiler import errors, external_types, internal_types, sentinels, utils


class VisitorMixinTemporalConstraints(PTNodeVisitor):
    """A visitor mixin for temporal constraints."""

    @staticmethod
    def visit_temporal_constraint(
        _, children: list[external_types.TemporalConstraint]
    ) -> external_types.TemporalConstraint:
        """Visit a <temporal_constraint> node."""
        return children[0]

    @staticmethod
    def visit_time_frame_statement(
        _, children: list[external_types.TimeFrameStatement]
    ) -> external_types.TimeFrameStatement:
        """Visit a <time_frame_statement> node."""
        return children[0]

    @staticmethod
    def visit_time_frame_statement_before(
        _, children: list[external_types.TimeDelta | internal_types.Sentinel | None]
    ) -> external_types.TimeFrameStatement:
        """Visit a <time_frame_statement_before> node."""
        frame_close, frame_anchor = children
        frame_close: external_types.TimeDelta
        frame_anchor: internal_types.Sentinel | None
        time_frame_statement = external_types.TimeFrameStatement(
            type=external_types.TemporalStatementDiscriminator.TIME_FRAME,
            open=None,
            close=frame_close,
            useActionTimestamp=frame_anchor is sentinels.ACTION_TEMPORAL_ANCHOR
        )
        return time_frame_statement

    @staticmethod
    def visit_time_frame_statement_after(
        _, children: list[external_types.TimeDelta | internal_types.Sentinel | None]
    ) -> external_types.TimeFrameStatement:
        """Visit a <time_frame_statement_after> node."""
        frame_open, frame_anchor = children
        frame_open: external_types.TimeDelta
        frame_anchor: internal_types.Sentinel | None
        time_frame_statement = external_types.TimeFrameStatement(
            type=external_types.TemporalStatementDiscriminator.TIME_FRAME,
            open=frame_open,
            close=None,
            useActionTimestamp=frame_anchor is sentinels.ACTION_TEMPORAL_ANCHOR
        )
        return time_frame_statement

    @staticmethod
    def visit_time_frame_statement_between(
        _, children: list[external_types.TimeDelta | internal_types.Sentinel | None]
    ) -> external_types.TimeFrameStatement:
        """Visit a <time_frame_statement_between> node."""
        frame_open, frame_close, frame_anchor = children
        frame_open: external_types.TimeDelta
        frame_close: external_types.TimeDelta
        frame_anchor: internal_types.Sentinel | None
        time_frame_statement = external_types.TimeFrameStatement(
            type=external_types.TemporalStatementDiscriminator.TIME_FRAME,
            open=frame_open,
            close=frame_close,
            useActionTimestamp=frame_anchor is sentinels.ACTION_TEMPORAL_ANCHOR
        )
        return time_frame_statement

    @staticmethod
    def visit_time_period(
        _, children: list[external_types.FloatField | external_types.IntField | external_types.TimeFrameTimeUnit]
    ) -> external_types.TimeDelta:
        """Visit a <time_period> node."""
        number_expression, time_unit = children
        number_expression: external_types.FloatField | external_types.IntField
        time_unit: external_types.TimeFrameTimeUnit
        number = number_expression['value']
        time_delta = external_types.TimeDelta(amount=number, unit=time_unit)
        return time_delta

    @staticmethod
    def visit_time_unit(node: NonTerminal, children: list[str]) -> external_types.TimeFrameTimeUnit:
        """Visit a <time_unit> node."""
        raw_time_unit = children[0]
        match raw_time_unit:
            case "minutes" | "hours" | "days" | "weeks" | "months" | "years":
                time_unit = raw_time_unit
            case "minute" | "hour" | "day" | "week" | "month" | "year":
                time_unit = raw_time_unit + 's'
            case _:
                raise errors.VivCompileError(
                    f"Unexpected time unit: {raw_time_unit}",
                    source=utils.derive_source_annotations(node=node)
                )
        return cast(external_types.TimeFrameTimeUnit, time_unit)

    @staticmethod
    def visit_time_frame_anchor(_, children: list[internal_types.Sentinel]) -> internal_types.Sentinel:
        """Visit a <time_frame_anchor> node."""
        return children[0]

    @staticmethod
    def visit_time_frame_anchor_from_action(_, __) -> internal_types.Sentinel:
        """Visit a <time_frame_anchor_from_action> node."""
        return sentinels.ACTION_TEMPORAL_ANCHOR

    @staticmethod
    def visit_time_frame_anchor_from_hearing(_, __) -> internal_types.Sentinel:
        """Visit a <time_frame_anchor_from_hearing> node."""
        return sentinels.NON_ACTION_TEMPORAL_ANCHOR

    @staticmethod
    def visit_time_frame_anchor_from_now(_, __) -> internal_types.Sentinel:
        """Visit a <time_frame_anchor_from_now> node."""
        return sentinels.NON_ACTION_TEMPORAL_ANCHOR

    @staticmethod
    def visit_time_frame_anchor_ago(_, __) -> internal_types.Sentinel:
        """Visit a <time_frame_anchor_ago> node."""
        return sentinels.NON_ACTION_TEMPORAL_ANCHOR

    @staticmethod
    def visit_time_of_day_statement(
        _, children: list[external_types.TimeOfDayStatement]
    ) -> external_types.TimeOfDayStatement:
        """Visit a <time_of_day_statement> node."""
        return children[0]

    @staticmethod
    def visit_time_of_day_statement_before(
        _, children: list[external_types.TimeOfDayDeclaration]
    ) -> external_types.TimeOfDayStatement:
        """Visit a <time_of_day_statement_before> node."""
        window_close = children[0]
        time_of_day_statement = external_types.TimeOfDayStatement(
            type=external_types.TemporalStatementDiscriminator.TIME_OF_DAY,
            open=None,
            close=window_close
        )
        return time_of_day_statement

    @staticmethod
    def visit_time_of_day_statement_after(
        _, children: list[external_types.TimeOfDayDeclaration]
    ) -> external_types.TimeOfDayStatement:
        """Visit a <time_of_day_statement_after> node."""
        window_open = children[0]
        time_of_day_statement = external_types.TimeOfDayStatement(
            type=external_types.TemporalStatementDiscriminator.TIME_OF_DAY,
            open=window_open,
            close=None
        )
        return time_of_day_statement

    @staticmethod
    def visit_time_of_day_statement_between(
        _, children: list[external_types.TimeOfDayDeclaration]
    ) -> external_types.TimeOfDayStatement:
        """Visit a <time_of_day_statement_between> node."""
        window_open, window_close = children
        time_of_day_statement = external_types.TimeOfDayStatement(
            type=external_types.TemporalStatementDiscriminator.TIME_OF_DAY,
            open=window_open,
            close=window_close
        )
        return time_of_day_statement

    @staticmethod
    def visit_time_of_day(node: NonTerminal, children: list[str]) -> external_types.TimeOfDayDeclaration:
        """Visit a <time_of_day> node."""
        raw_hour = int(children[0])
        twelve_hour_form = children[-1] in ("am", "pm")
        if twelve_hour_form:
            period = children[-1]
            minute = int(children[1]) if len(children) > 2 else 0
            if not 1 <= raw_hour <= 12:
                raise errors.VivCompileError(
                    f"Invalid hour in 12-hour time: {raw_hour}",
                    source=utils.derive_source_annotations(node=node)
                )
            hour = (raw_hour % 12) + (12 if period == "pm" else 0)
        else:
            minute = int(children[1]) if len(children) > 1 else 0
            if not 0 <= raw_hour <= 23:
                raise errors.VivCompileError(
                    f"Invalid hour in 24-hour time: {raw_hour}",
                    source=utils.derive_source_annotations(node=node)
                )
            hour = raw_hour
        if not 0 <= minute <= 59:
            raise errors.VivCompileError(f"Invalid minute in time: {minute}", source=utils.derive_source_annotations(node=node))
        return external_types.TimeOfDayDeclaration(hour=hour, minute=minute)

    @staticmethod
    def visit_time_of_day_hh(node: NonTerminal, _) -> str:
        """Visit a <time_of_day_hh> node."""
        return str(node)

    @staticmethod
    def visit_time_of_day_mm(node: NonTerminal, _) -> str:
        """Visit a <time_of_day_mm> node."""
        return str(node)



"""Handles importing between Viv files."""

__all__ = ["integrate_included_files"]

from pathlib import Path

import arpeggio
from arpeggio.cleanpeg import ParserPEG

from viv_compiler import errors, internal_types, utils, validation, visitor


def integrate_included_files(
    viv_parser: ParserPEG,
    viv_visitor: visitor.Visitor,
    ast: internal_types.AST,
    entry_point_file_path: Path,
) -> internal_types.CombinedAST:
    """Handle any `include` declarations in the given AST (including any recursive ones)
    and return a dictionary containing trope definitions and action definitions.

    Args:
        viv_parser: A prepared Viv parser.
        viv_visitor: A prepared Viv AST visitor.
        ast: An abstract syntax tree produced by the Visitor class.
        entry_point_file_path: The file path for the file that is being directly compiled by the author.

    Returns:
        A combined AST containing content compiled from the entry-point file along
        with that of any included files.
    """
    # First, we need to recursively gather all included files. If we do this step first, we don't
    # need to worry about circular dependencies (Viv allows circular imports).
    all_included_asts: list[internal_types.AST] = [ast]
    _parse_all_included_files(
        viv_parser=viv_parser,
        viv_visitor=viv_visitor,
        ast=ast,
        anchor_dir=entry_point_file_path.parent,
        included_file_paths={entry_point_file_path},
        included_file_asts=all_included_asts
    )
    # We'll start with the AST for the user's entry file. If there's no includes,
    # will end up being the combined AST.
    combined_ast = internal_types.CombinedAST(
        actions=[],
        actionSelectors=[],
        plans=[],
        planSelectors=[],
        queries=[],
        siftingPatterns=[],
        tropes=[]
    )
    # Now let's parse each of the included files to build up a combined AST
    for included_ast in all_included_asts:
        _integrate_included_file(combined_ast=combined_ast, included_file_ast=included_ast)
    # Ensure that we do not have any duplicate names (either within a file or across files)
    validation.prevalidate_construct_names(combined_ast=combined_ast)
    # Return the combined AST
    return combined_ast


def _parse_all_included_files(
    viv_parser: ParserPEG,
    viv_visitor: visitor.Visitor,
    ast: internal_types.AST,
    anchor_dir: Path,
    included_file_paths: set[Path],
    included_file_asts: list[internal_types.AST]
) -> None:
    """Parse all the files included in the one at hand, and so forth recursively.

    Critically, this method is robust to circular includes, however arcane, and it captures
    arbitrarily recursive includes.

    Args:
        viv_parser: A prepared Viv parser.
        viv_visitor: A prepared Viv AST visitor.
        ast: An abstract syntax tree produced by the Visitor class.
        anchor_dir: The directory from which relative include paths will be resolved.
        included_file_paths: A running set containing absolute paths for all included
            files parsed so far. This will be mutated in place.
        included_file_asts: A running list containing ASTs for all included files parsed
            so far. This will also be mutated in place.

    Returns:
        Nothing. Mutates both `included_file_paths` and `included_file_asts` in place.

    Raises:
        VivCompileError: An included file does not exist.
        VivParseError: An included file could not be parsed.
    """
    for included_file_relative_path in ast["_includes"]:
        included_file_absolute_path = (anchor_dir / included_file_relative_path).resolve()
        if included_file_absolute_path in included_file_paths:  # Already captured this one
            continue
        included_file_paths.add(included_file_absolute_path)
        try:
            included_file_contents = included_file_absolute_path.read_text(encoding="utf-8")
        except FileNotFoundError:
            error_message = f"Bad 'include' declaration (file not found): {included_file_relative_path}"
            raise errors.VivCompileError(error_message)
        utils.register_source_file(
            source_file_path=included_file_absolute_path,
            source_file_contents=included_file_contents,
            position_to_line_and_column=viv_parser.pos_to_linecol
        )
        try:
            included_file_parse_tree = viv_parser.parse(_input=included_file_contents)
        except arpeggio.NoMatch as parsing_error:
            raise errors.VivParseError(original=parsing_error, file_path=included_file_absolute_path) from None
        included_file_ast: internal_types.AST = utils.sanitize_ast(
            ast=arpeggio.visit_parse_tree(parse_tree=included_file_parse_tree, visitor=viv_visitor)
        )
        included_file_asts.append(included_file_ast)
        _parse_all_included_files(
            viv_parser=viv_parser,
            viv_visitor=viv_visitor,
            ast=included_file_ast,
            anchor_dir=included_file_absolute_path.parent,
            included_file_paths=included_file_paths,
            included_file_asts=included_file_asts
        )


def _integrate_included_file(combined_ast: internal_types.CombinedAST, included_file_ast: internal_types.AST) -> None:
    """Integrate into the given combined AST the constructs in the given AST for an included file.

    Args:
        combined_ast: The combined AST that is currently being constructed.
        included_file_ast: AST for the file whose constructs will be integrated into the combined AST.

    Returns:
        Nothing. The combined AST is mutated in place.
    """
    combined_ast["actions"] += included_file_ast["actions"]
    combined_ast["actionSelectors"] += included_file_ast["actionSelectors"]
    combined_ast["plans"] += included_file_ast["plans"]
    combined_ast["planSelectors"] += included_file_ast["planSelectors"]
    combined_ast["queries"] += included_file_ast["queries"]
    combined_ast["siftingPatterns"] += included_file_ast["siftingPatterns"]
    combined_ast["tropes"] += included_file_ast["tropes"]
